@@ -147,6 +147,88 @@ exo_mount_hal_propagate_error (GError   **error,
 
 
 
+static gint
+exo_mount_hal_udi_is_volume (const gchar  *udi,
+                             GError      **error)
+{
+  gchar     **interfaces;
+  DBusError   derror;
+  guint       n;
+  gint        is_volume = -1;
+
+  g_return_val_if_fail (udi != NULL, -1);
+  g_return_val_if_fail (error == NULL || *error == NULL, -1);
+
+  /* initialize D-Bus error */
+  dbus_error_init (&derror);
+
+  /* determine the info.interfaces property of the device */
+  interfaces = libhal_device_get_property_strlist (hal_context, udi, "info.interfaces", &derror);
+  if (G_LIKELY (interfaces != NULL))
+    {
+      /* verify that we have a mountable device here */
+      for (is_volume = 0, n = 0; is_volume == 0 && interfaces[n] != NULL; ++n)
+        if (strcmp (interfaces[n], "org.freedesktop.Hal.Device.Volume") == 0)
+          is_volume = 1;
+      libhal_free_string_array (interfaces);
+    }
+  else
+    {
+      exo_mount_hal_propagate_error (error, &derror);
+    }
+
+  return is_volume;
+}
+
+
+
+static gchar *
+exo_mount_hal_get_volume_udi (const gchar  *udi,
+                              GError      **error)
+{
+  DBusError  derror;
+  gint       result;
+  gchar     *volume_udi = NULL;
+  gchar    **parent_udis;
+  gint       n, n_parent_udis;
+
+  g_return_val_if_fail (udi != NULL, NULL);
+  g_return_val_if_fail (error == NULL || *error == NULL, NULL);
+
+  result = exo_mount_hal_udi_is_volume (udi, error);
+  if (G_LIKELY (result == 1))
+    {
+      /* this udi is a valid volume */
+      volume_udi = g_strdup (udi);
+    }
+  else if (result == 0)
+    {
+      /* initialize D-Bus error */
+      dbus_error_init (&derror);
+
+      /* ok, but maybe we have a volume whose parent is identified by the udi */
+      parent_udis = libhal_manager_find_device_string_match (hal_context, "info.parent", udi, &n_parent_udis, &derror);
+      if (G_LIKELY (parent_udis != NULL))
+        {
+          for (n = 0; n < n_parent_udis; n++)
+            {
+              /* check if this udi is a volume, or one of its parents */
+              volume_udi = exo_mount_hal_get_volume_udi (parent_udis[n], error);
+              if (volume_udi != NULL || *error != NULL)
+                break;
+            }
+        }
+      else
+        {
+          exo_mount_hal_propagate_error (error, &derror);
+        }
+    }
+
+  return volume_udi;
+}
+
+
+
 /**
  * exo_mount_hal_device_from_udi:
  * @udi   : UDI of a volume or drive.
@@ -164,11 +246,7 @@ exo_mount_hal_device_from_udi (const gchar *udi,
 {
   ExoMountHalDevice *device = NULL;
   DBusError          derror;
-  gchar            **interfaces;
-  gchar            **volume_udis;
-  gchar             *volume_udi = NULL;
-  gint               n_volume_udis;
-  gint               n;
+  gchar             *volume_udi;
   gchar             *key;
 
   g_return_val_if_fail (udi != NULL, NULL);
@@ -178,61 +256,20 @@ exo_mount_hal_device_from_udi (const gchar *udi,
   if (!exo_mount_hal_init (error))
     return NULL;
 
-  /* initialize D-Bus error */
-  dbus_error_init (&derror);
-
-again:
-  /* determine the info.interfaces property of the device */
-  interfaces = libhal_device_get_property_strlist (hal_context, udi, "info.interfaces", &derror);
-  if (G_UNLIKELY (interfaces == NULL))
+  /* find the volume udi for this udi (either this udi or one of its parents) */
+  volume_udi = exo_mount_hal_get_volume_udi (udi, error);
+  if (G_UNLIKELY (volume_udi == NULL))
     {
-      /* reset D-Bus error */
-      dbus_error_free (&derror);
-
-      /* release any previous volume UDI */
-      g_free (volume_udi);
-      volume_udi = NULL;
-
-      /* ok, but maybe we have a volume whose parent is identified by the udi */
-      volume_udis = libhal_manager_find_device_string_match (hal_context, "info.parent", udi, &n_volume_udis, &derror);
-      if (G_UNLIKELY (volume_udis == NULL))
-        {
-err0:     exo_mount_hal_propagate_error (error, &derror);
-          goto out;
-        }
-      else if (G_UNLIKELY (n_volume_udis < 1))
-        {
-          /* no match, we cannot handle that device */
-          libhal_free_string_array (volume_udis);
-          goto err1;
-        }
-
-      /* use the first volume UDI... */
-      volume_udi = g_strdup (volume_udis[0]);
-      libhal_free_string_array (volume_udis);
-
-      /* ..and try again using that UDI */
-      udi = (const gchar *) volume_udi;
-      goto again;
-    }
-
-  /* verify that we have a mountable device here */
-  for (n = 0; interfaces[n] != NULL; ++n)
-    if (strcmp (interfaces[n], "org.freedesktop.Hal.Device.Volume") == 0)
-      break;
-  if (G_UNLIKELY (interfaces[n] == NULL))
-    {
-      /* definitely not a device that we're able to mount, eject or unmount */
-err1: g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_FAILED, _("Given device \"%s\" is not a volume or drive"), udi);
-      goto out;
+err0: g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_FAILED, _("Given device \"%s\" is not a volume or drive"), udi);
+      return NULL;
     }
 
   /* setup the device struct */
   device = g_new0 (ExoMountHalDevice, 1);
-  device->udi = g_strdup (udi);
+  device->udi = volume_udi;
 
   /* check if we have a volume here */
-  device->volume = libhal_volume_from_udi (hal_context, udi);
+  device->volume = libhal_volume_from_udi (hal_context, device->udi);
   if (G_LIKELY (device->volume != NULL))
     {
       /* determine the storage drive for the volume */
@@ -245,14 +282,14 @@ err1: g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_FAILED, _("Given device \"%
 
           /* setup the file system internals */
           device->fstype = libhal_volume_get_fstype (device->volume);
-          device->altfstype = libhal_device_get_property_string (hal_context, udi, "volume.fstype.alternative.preferred", NULL);
+          device->altfstype = libhal_device_get_property_string (hal_context, device->udi, "volume.fstype.alternative.preferred", NULL);
           device->fsusage = libhal_volume_get_fsusage (device->volume);
         }
     }
   else
     {
       /* check if we have a drive here (i.e. floppy) */
-      device->drive = libhal_drive_from_udi (hal_context, udi);
+      device->drive = libhal_drive_from_udi (hal_context, device->udi);
       if (G_LIKELY (device->drive != NULL))
         {
           /* setup the device internals */
@@ -265,23 +302,25 @@ err1: g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_FAILED, _("Given device \"%
         }
     }
 
+  /* initialize D-Bus error */
+  dbus_error_init (&derror);
+
   /* determine the valid mount options from the UDI */
   if (device->altfstype != NULL)
     {
       key = g_strdup_printf ("volume.mount.%s.valid_options", device->altfstype);
-      device->fsoptions = libhal_device_get_property_strlist (hal_context, udi, key, &derror);
+      device->fsoptions = libhal_device_get_property_strlist (hal_context, device->udi, key, &derror);
       g_free (key);
     }
   else
     {
-      device->fsoptions = libhal_device_get_property_strlist (hal_context, udi, "volume.mount.valid_options", &derror);
+      device->fsoptions = libhal_device_get_property_strlist (hal_context, device->udi, "volume.mount.valid_options", &derror);
     }
 
   /* sanity checking */
   if (G_UNLIKELY (device->file == NULL || device->name == NULL))
     {
       exo_mount_hal_device_free (device);
-      device = NULL;
       goto err0;
     }
 
@@ -289,16 +328,11 @@ err1: g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_FAILED, _("Given device \"%
   if (G_LIKELY (device->drive == NULL))
     {
       /* definitely not a device that we're able to mount, eject or unmount */
-      g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_FAILED, _("Given device \"%s\" is not a volume or drive"), udi);
+      g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_FAILED, _("Given device \"%s\" is not a volume or drive"), device->udi);
       exo_mount_hal_device_free (device);
       device = NULL;
     }
 
-out:
-  /* cleanup */
-  libhal_free_string_array (interfaces);
-  g_free (volume_udi);
-
   return device;
 }
 
@@ -660,8 +694,14 @@ exo_mount_hal_device_mount (ExoMountHalDevice *device,
   gchar       *fstype;
   const gchar *fs;
   gchar       *s;
-  gint         m, n = 0;
+  guint        m, n;
   const gchar *charset;
+  XfceRc      *rc;
+  gchar       *key;
+  gchar       *option;
+  GSList      *lp, *fsoptions = NULL;
+  gsize        len;
+  const gchar *value;
 
   g_return_val_if_fail (device != NULL, FALSE);
   g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
@@ -676,60 +716,98 @@ exo_mount_hal_device_mount (ExoMountHalDevice *device,
   else
     fs = device->fstype;
 
+  /* open the config file and look for the filesystem group */
+  rc = xfce_rc_config_open (XFCE_RESOURCE_CONFIG, "xfce4" G_DIR_SEPARATOR_S "mount.rc", FALSE);
+  if (G_LIKELY (rc != NULL))
+    {
+      if (xfce_rc_has_group (rc, fs))
+        {
+          /* set the filesystem group */
+          xfce_rc_set_group (rc, fs);
+        }
+      else
+        {
+          /* config is not usable */
+          xfce_rc_close (rc);
+          rc = NULL;
+        }
+    }
+
   /* check if we know any valid mount options */
   if (G_LIKELY (device->fsoptions != NULL))
     {
       /* process all valid mount options */
       for (m = 0; device->fsoptions[m] != NULL; ++m)
         {
-          /* this is currently mostly Linux specific noise */
-          if (strcmp (device->fsoptions[m], "uid=") == 0
-              && (strcmp (fs, "vfat") == 0
-               || strcmp (fs, "iso9660") == 0
-               || strcmp (fs, "udf") == 0
-               || strcmp (fs, "ntfs") == 0
-               || strcmp (fs, "ntfs-3g") == 0
-               || device->volume == NULL))
-            {
-              options[n++] = g_strdup_printf ("uid=%u", (guint) getuid ());
-            }
-          else if (strcmp (device->fsoptions[m], "shortname=") == 0
-                && strcmp (fs, "vfat") == 0)
-            {
-              options[n++] = g_strdup_printf ("shortname=winnt");
-            }
-          else if (strcmp (device->fsoptions[m], "sync") == 0
-                && device->volume == NULL)
+          option = NULL;
+
+          if (strcmp (device->fsoptions[m], "sync") == 0)
             {
               /* non-pollable drive... */
-              options[n++] = g_strdup ("sync");
+              option = g_strdup ("sync");
             }
-          else if (strcmp (device->fsoptions[m], "longnames") == 0
-                && strcmp (fs, "vfat") == 0)
+          else if (rc != NULL)
             {
-              /* however this one is FreeBSD specific */
-              options[n++] = g_strdup ("longnames");
-            }
-          else if (strcmp (device->fsoptions[m], "umask=") == 0
-                   && strcmp (fs, "ntfs-3g") == 0)
-            {
-              /* we need to pass umask=0077 to ntfs-g3 or else it gets 0777 perms */
-              options[n++] = g_strdup ("umask=0077");
-            }
-          else if (strcmp (device->fsoptions[m], "iocharset=") == 0)
-            {
-              /* get the charset from a variable set by the user or glib */
-              charset = g_getenv ("EXO_MOUNT_IOCHARSET");
-              if (G_LIKELY (charset == NULL))
-                if (g_get_charset (&charset))
-                  charset = "utf8";
+              /* option with value or enabled/disabled */
+              if (g_str_has_suffix (device->fsoptions[m], "="))
+                {
+                  len = strlen (device->fsoptions[m]) - 1;
+                  key = g_strndup (device->fsoptions[m], len);
+                  value = xfce_rc_read_entry_untranslated (rc, key, NULL);
 
-              if (G_LIKELY (charset != NULL && *charset != '\0'))
-                options[n++] = g_strdup_printf ("iocharset=%s", charset);
+                  if (value != NULL)
+                    {
+                      /* substitute the <auto> options */
+                      if (strcmp (value, "<auto>") == 0)
+                        {
+                          if (strcmp (key, "uid") == 0)
+                            {
+                              option = g_strdup_printf ("uid=%u", (guint) getuid ());
+                            }
+                          else if (strcmp (key, "gid") == 0)
+                            {
+                              option = g_strdup_printf ("gid=%u", (guint) getgid ());
+                            }
+                          else if (strcmp (key, "iocharset") == 0)
+                            {
+                              charset = g_getenv ("EXO_MOUNT_IOCHARSET");
+                              if (charset == NULL
+                                  && g_get_charset (&charset))
+                                charset = "utf8";
+                              option = g_strdup_printf ("iocharset=%s", charset);
+                            }
+                        }
+                      else
+                        {
+                          /* use the value from the rc file */
+                          option = g_strdup_printf ("%s=%s", key, value);
+                        }
+                    }
+
+                  g_free (key);
+                }
+              else if (xfce_rc_has_entry (rc, device->fsoptions[m]))
+                {
+                  if (xfce_rc_read_bool_entry (rc, device->fsoptions[m], FALSE))
+                    option = g_strdup (device->fsoptions[m]);
+                }
             }
+
+          /* add the option */
+          if (option != NULL)
+            fsoptions = g_slist_prepend (fsoptions, option);
         }
     }
 
+  if (rc != NULL)
+    xfce_rc_close (rc);
+
+  /* create the filesystem options (+2 for possible "ro" and null terminate) */
+  options = g_new0 (gchar *, g_slist_length (fsoptions) + 2);
+  for (n = 0, lp = fsoptions; lp != NULL; lp = lp->next, ++n)
+    options[n] = lp->data;
+  g_slist_free (fsoptions);
+
   /* try to determine a usable mount point */
   if (G_LIKELY (device->volume != NULL))
     {
@@ -743,9 +821,10 @@ exo_mount_hal_device_mount (ExoMountHalDevice *device,
     }
 
   /* make sure that the mount point is usable (i.e. does not contain G_DIR_SEPARATOR's) */
-  mount_point = (mount_point != NULL && *mount_point != '\0')
-              ? exo_str_replace (mount_point, G_DIR_SEPARATOR_S, "_") 
-              : g_strdup ("");
+  if (mount_point != NULL && *mount_point != '\0')
+    mount_point = exo_str_replace (mount_point, G_DIR_SEPARATOR_S, "_");
+  else
+    mount_point = g_strdup ("");
 
   /* let HAL guess the fstype, unless we have an alternative preferred fstype */
   if (G_UNLIKELY (device->altfstype != NULL))
@@ -879,7 +958,7 @@ oom:      g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_NOMEM, "%s", g_strerror
   if (dbus_error_is_set (&derror))
     {
       /* try to translate the error appropriately */
-      if (strcmp (derror.name, "org.freedesktop.Hal.Device.Volume.PermissionDenied") == 0) 
+      if (strcmp (derror.name, "org.freedesktop.Hal.Device.Volume.PermissionDenied") == 0)
         {
           /* TRANSLATORS: User tried to mount a volume, but is not privileged to do so. */
           g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_FAILED, _("You are not privileged to mount the volume \"%s\""), device->name);
@@ -1059,7 +1138,7 @@ oom:  g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_NOMEM, "%s", g_strerror (EN
   if (G_UNLIKELY (dbus_error_is_set (&derror)))
     {
       /* try to translate the error appropriately */
-      if (strcmp (derror.name, "org.freedesktop.Hal.Device.Volume.PermissionDenied") == 0) 
+      if (strcmp (derror.name, "org.freedesktop.Hal.Device.Volume.PermissionDenied") == 0)
         {
           /* TRANSLATORS: User tried to unmount a volume, but is not privileged to do so. */
           g_set_error (error, G_FILE_ERROR, G_FILE_ERROR_FAILED, _("You are not privileged to unmount the volume \"%s\""), device->name);
